/*****************************************************************************/
/*
                                 CGIutl.c

A CLI utility to assist at the DCL level with generating HTTP responses,
returning content and with the handling of POSTed requests.  Intended to be
used within DCL procedures to manipulate the CGI environment.

Assign a foreign verb,

  $ CGIUTL = "$HT_EXE:CGIUTL"

All DCL symbols created by or supplied to this utility MUST be global symbols. 
It does not deal with local symbols at all.


DECODING A REQUEST BODY TO DCL SYMBOLS
--------------------------------------
The /URLDECODE qualifier results in a form-URL-encoded POSTed request body
being decoded and the contents of each field being placed into DCL symbols
named using that field.  HTML form field names may contain characters not
allowed within DCL symbol names, therefore when constructing the symbol name
non-alpha-numeric characters in the field name have underscores substituted in
the symbol name.  Once the request body has been decoded the DCL procedure can
manipulate the contents quite simply.  Here's the command line:

  $ CGIUTL /URLDECODE

For example the following form input field

  <INPUT TYPE=text NAME=text_input_one SIZE=40>

would have the corresponding DCL symbol name assigned

  CGIUTL_TEXT_INPUT_ONE == "blah-blah-blah-blah"

Field contents greater than the capacity of a single DCL symbol (255 on <=6.n
and 1023 on >=7.n) have the contents distributed across multiple symbol names.
Each of these fields comprises the basic name with a dollar symbol and an
ascending integer.  A symbol with the integer zero contains the total count of
the contents of all of the symbols.  For example, the following text-area field
may have had 900 characters entered.

  <TEXTAREA NAME=text_area_input ROWS=14 COLS=80>
  </TEXTAREA>

Assuming a maximum symbol capacity of 255 characters the contents would be
distributed across 4 symbols, with the additional count symbol, as follows:

  CGIUTL_TEXT_AREA_INPUT$0 == "900"
  CGIUTL_TEXT_AREA_INPUT$1 == "blah-blah-blah-blah...blah-blah"
  CGIUTL_TEXT_AREA_INPUT$2 == "blah-blah-blah-blah...blah-blah"
  CGIUTL_TEXT_AREA_INPUT$3 == "blah-blah-blah-blah...blah-blah"
  CGIUTL_TEXT_AREA_INPUT$4 == "blah-blah-blah"

In addition three special-purpose symbol provide additional information about
the DCL symbols created.

  CGIUTL$FIELDS .... the number of fields in the body (and hence <FORM>)

  CGIUTL$MAXSYM .... the maximum number of characters possible in each symbol

  CGIUTL$SYMBOLS ... the number of field symbols created by the utility
                     (this excludes these three special-purpose symbols)
                     if a simple comparision between this and CGIUTL$FIELDS
                     show a different count then at least one field had to
                     be distributed across mulitple symbols due to size


DECODING A REQUEST BODY TO A FILE
---------------------------------
The /URLDECODE used with the /OUTPUT=filename qualifier results in a
form-URL-encoded POSTed request body being written to the specified file.  Two
formats are possible controlled using the /FORMAT=NAMES (default) and
/FORMAT=HEADINGS qualifier and keywords.  When NAMES are used each field name
of the format is followed by a colon, space and then the field value.  With
HEADINGS the field name is on a line by itself, underlined with hyphens and
then in a block the field values (more suited to <TEXTAREA> input).  To write
to an output file as well as create DCL symbols add the /SYMBOLS qualifier. 
The following examples show some of the variations.

  $ CGIUTL /URLDECODE /OUTPUT=HT_ROOT:[LOG.SERVER]CGIUTL_FORM.TMP
  $ CGIUTL /URLDECODE /OUTPUT=HT_ROOT:[LOG.SERVER]CGIUTL.TMP /FORMAT=HEADINGS
  $ CGIUTL /URLDECODE /OUTPUT=HT_ROOT:[LOG.SERVER]CGIUTL_FORM.TMP /SYMBOLS


WRITING A REQUEST BODY TO A FILE
--------------------------------
Using the /BODY qualifier any POSTed request body (form-URL-encoded, text or
other content) can simply be written to the specified /OUTPUT= file.  It will
be created in STREAM-LF format.  For form-URL-encoded and text/.. bodies the
file is written as text.  For other content-type the file is written as binary.


NOTE ON FORM FIELDS
-------------------
Most form fields produce a name in the form data stream (URL-encoded request
body) regardless of whether they are empty or not.  Checkbox fields do not. 
Unless checkboxes are checked the associated names and values do not appear in
the data stream and will not have DCL symbols created for them.  Three possible
approaches:

1. Initialize symbols related to checkbox field names to empty values prior to
using CGIUTL.  This way the symbol can always be guaranteed to exist in
subsequent processing, overwritten with the form value if checked.

2. After using CGIUTL use F$TYPE() to check if an associated symbol exists.

3. Include a HIDDEN INPUT field before checkboxes to ensure the relevant field
name always exists in the data stream.  Example:

  <INPUT TYPE=hidden NAME=testing VALUE="">
  <INPUT TYPE=checkbox NAME=testing VALUE=1> Just Testing

Incorrect and/or malicious form POSTings should also be allowed for, where
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ expected field names may be absent,
disrupting script processing.  Accomodating such possibilities suggests
adopting a strategy of either initializing all expected symbols names BEFORE
calling CGIUTL, or testing for the existance of all expected symbol names AFTER
calling CGIUTL.


MASSAGING DCL SYMBOLS
---------------------
The utility will adjust the number of double-quotes in a DCL symbol making it
suitable for substitution, parameter passing, etc., anywhere a single quote in
a symbol would cause problems.  This may be done multiple times, first
doubling, then quadrupling, etc., the number of quotes.  Buffer or symbol
overflow is of course possible.  These are errors.  If the symbol does not
exist or is a local symbol an error is reported.  Example:

  $ CGIUTL THIS_GLOBAL_SYMBOL /2QUOTES


GENERATING AN HTTP RESPONSE
---------------------------
The utility can be used to provide a response header and/or body.

The /RESPONSE=[status-code] qualifier generates an "HTTP/1.0 status-code
string" HTTP response.  The default is "200 Success".

If the /CONTENT=string qualifier is applied the specified MIME content-type
header line is generated.  Otherwise content-type defaults to "text/plain".

If the /RESPONSE qualifier is used with a file name parameter, a response
header is generated and then the contents of the file are transfered to the
client as the request body.  The file is either opened in record mode if the
content-type is "text/..." (by default, without a /CONTENT) or in binary mode
if any other content-type.  For example:

  $ CGIUTL /RESPONSE=404 WEB:[ERRORS]NOTFOUND.HTML
  $ CGIUTL /RESPONSE HT_ROOT:[000000]HTTPDWASD.GIF /CONTENT="image/gif"

If /LENGTH is used with /RESPONSE when specifying a file name, as above, the
file is first completely read to determine the actual content length (not
fstat()ed, which does not work with variable-record-length files) and this
added to the response header, before rewind()ing and transfering the file
contents.  For example:

  $ CGIUTL /RESPONSE /LENGTH HT_ROOT:[000000]README.TXT

If /COPY is used _without_ /RESPONSE against a parameter file name the content
is copied to the client as above _without_ first sending a response header
(i.e. the script must provide it's own).  This is one way to transfer a binary
file to a client from DCL.  Note that the /CONTENT= is required so that it can
be determined whether it is text or binary content.  Example:

  $ SAY "HTTP/1.0 200 Success" + LF
  $ SAY "Content-Type: image/gif" + LF
  $ SAY LF
  $ CGIUTL /COPY HT_ROOT:[000000]HTTPDWASD.GIF /CONTENT="image/gif"

Redirection responses may be generated using the /LOCATION=url-path qualifier. 
If this is supplied the status code is forced to 302.  An example:

  $ CGIUTL /LOCATION="http://www.openvms.digital.com/"


QUALIFIERS
----------
/2QUOTE              double up the quotes in the parameter symbol
/BODY                output the request body
/CONTENT=string      response content-type (default "text/plain")
/COPY                copy parameter file to client
/DBUG                turns on all "if (Debug)" statements
/EXPIRED             response header has an "Expired:" pre-expired time added
/FORMAT=             when writing form-URL-encoded body as file
                     "HEADINGS" as underlined headings
                     "NAMES" as "field-name: field-value"
/LOCATION=string     send a 302 HTTP header with this as the location
/MAXSYM=integer      maximum number of characters per symbol value
/OUTPUT=file         output to specified file
/PREFIX=string       prefix to symbol name (e.g. "WWW", defaults to "CGIUTL")
/RESPONSE[=integer]  send response header with this status code (default 200)
/SYMBOLS             put into DCL symbols
/URLDECODE           decode a form-URL-encoded, POSTed body
                     (use with /SYMBOL and/or /OUTPUT=)


LOGICAL NAMES
-------------
CGIUTL$DBUG        turns on all "if (Debug)" statements
CGIUTL$PARAM       equivalent to (overrides) the command line
                   parameters/qualifiers (define as a system-wide logical)
                   may also be assigned as a local/global symbol


BUILD DETAILS
-------------
See BUILD_ONE.COM procedure.


COPYRIGHT
---------
Copyright (c) 1996-1999 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY (update 'SoftwareID' as well)
---------------
13-FEB-1999  MGD  v1.1.0, response generating functionality
06-FEB-1999  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "CGIUTL AXP-1.1.0";
#else
   char SoftwareID [] = "CGIUTL VAX-1.1.0";
#endif

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>

#include <descrip.h>
#include <libdef.h>
#include <libclidef.h>
#include <syidef.h>
#include <ssdef.h>
#include <stsdef.h>

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define boolean int
#define true 1
#define false 0

/* macro provides NULL pointer if CGI variable does not exist */
#define GetCgiVarIfExists(CharPointer,CgiVariableName) \
   CharPointer = getenv(CgiVariableName)

/* macro provides pointer to empty string even if CGI variable does not exist */
#define GetCgiVar(CharPointer,CgiVariableName) \
{ \
   if (CharPointer == NULL) \
      if ((CharPointer = getenv(CgiVariableName)) == NULL) \
          CharPointer = ""; \
   if (Debug) fprintf (stdout, "%s |%s|\n", CgiVariableName, CharPointer); \
}

#define FORM_URLENCODED "application/x-www-form-urlencoded"
#define BUFFER_CHUNK 1024

char  Utility [] = "CGIUTL";

boolean  Debug,
         DoContentLength,
         DoCopy,
         DoDclSymbols,
         DoEscapeSymbolQuotes,
         DoFormatHeadings,
         DoFormatNames,
         DoPreExpired,
         DoRequestBody,
         DoResponse,
         DoUrlDecode,
         ResponseContentTypeText;

int  BufferCount,
     BufferSize,
     FileContentLength,
     ReadCount,
     SymbolValueMax,
     SymbolPrefixLength;

char  *BufferPtr,
      *CgiContentTypePtr,
      *CgiRequestMethodPtr,
      *CgiRequestTimeGmtPtr,
      *CgiServerSoftwarePtr,
      *CharsetPtr,
      *CliCharsetPtr,
      *CliContentTypePtr,
      *CliLocationPtr,
      *CliOutputPtr,
      *CliParameterPtr,
      *CliResponsePtr,
      *SymbolPrefixPtr;

char  CharsetString [256];

/* required prototypes */
char* HttpStatusCodeText (int);

/*****************************************************************************/
/*
*/

main ()

{
   register char  *cptr;

   boolean  Done;

   /*********/
   /* begin */
   /*********/

   if (getenv ("CGIUTL$DBUG") != NULL) Debug = true;

   GetParameters ();

   if (CliOutputPtr != NULL)
      if (!CliOutputPtr[0])
         CliOutputPtr = CliParameterPtr;

   if (VmsVersion() >= 7)
   {
      if (SymbolValueMax < 1 || SymbolValueMax > 1023)
         SymbolValueMax = 1023;
   }
   else
   {
      if (SymbolValueMax < 1 || SymbolValueMax > 255)
         SymbolValueMax = 255;
   }
   if (Debug) fprintf (stdout, "SymbolValueMax: %d\n", SymbolValueMax);

   if (SymbolPrefixPtr == NULL)
   {
      SymbolPrefixPtr = Utility;
      SymbolPrefixLength = strlen(Utility);
   }

   Done = false;

   if (DoEscapeSymbolQuotes)
   {
      /****************************/
      /* escape a symbol's quotes */
      /****************************/

      if (CliParameterPtr == NULL)
      {
         fprintf (stdout, "%%%s-E-2QUOTE, symbol name not specified\n",
                  Utility);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      EscapeSymbolQuotes (CliParameterPtr);
      exit (SS$_NORMAL);
   }

   if (DoResponse || CliParameterPtr != NULL)
   {
      /*********************/
      /* generate response */
      /*********************/

      GetCgiVar (CgiServerSoftwarePtr, "WWW_SERVER_SOFTWARE");
      if (!CgiServerSoftwarePtr[0]) CgiServerSoftwarePtr = SoftwareID;

      GetCgiVar (CgiRequestTimeGmtPtr, "WWW_REQUEST_TIME_GMT");

      if (CliContentTypePtr == NULL)
      {
         CliContentTypePtr = "text/plain";
         ResponseContentTypeText = true;
      }
      else
      {
         /* ensure it's lower case */
         for (cptr = CliContentTypePtr; *cptr; cptr++)
            *cptr = tolower(*cptr);
         ResponseContentTypeText = !memcmp (CliContentTypePtr, "text/", 5);
      }

      if (ResponseContentTypeText)
      {
         GetCgiVar (CharsetPtr, "WWW_SERVER_CHARSET");
         if (!CharsetPtr[0]) CharsetPtr = CliCharsetPtr;
         if (CharsetPtr == NULL) CharsetPtr = "ISO-8859-1";
         sprintf (CharsetString, "; charset=%s", CharsetPtr);
         CharsetPtr = CharsetString;
      }
      else
         CharsetPtr = "";

      if (CliResponsePtr == NULL) CliResponsePtr = "200";

      FileContentLength = -1;

      if (DoCopy && CliParameterPtr == NULL)
      {
         fprintf (stdout, "%%%s-E-COPY, file not specified\n", Utility);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      if (CliParameterPtr == NULL)
         SendResponse ();
      else
         SendFile ();

      exit (SS$_NORMAL);
   }

   /**************************/
   /* process POSTed request */
   /**************************/

   if (DoUrlDecode && DoDclSymbols)
   {
      /* form-URL-decode the body and create corresponding DCL symbols */
      UrlDecodeBodyDclSymbols ();
      Done = true;
   }

   if (DoUrlDecode && CliOutputPtr != NULL)
   {
      /* form-URL-decode the body and write to file as field name/values */
      UrlDecodeBodyWrite ();
      Done = true;
   }

   if (!DoUrlDecode && CliOutputPtr != NULL)
   {
      /* write the request body to file as plain text */
      RequestBodyWrite ();
      Done = true;
   }

   if (!Done)
   {
      fprintf (stdout, "%%%s-E-HUH, did nothing!\n", Utility);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   register char  *aptr, *cptr, *clptr, *sptr;

   int  status;
   unsigned short  Length;
   char  ch;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("CGIUTL$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (strsame (aptr, "/2QUOTE", 3))
      {
         DoEscapeSymbolQuotes = true;
         continue;
      }
      if (strsame (aptr, "/BODY", 4))
      {
         DoRequestBody = true;
         continue;
      }
      if (strsame (aptr, "/CHARSET=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliCharsetPtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/CONTENT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliContentTypePtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/COPY", 4))
      {
         DoCopy = true;
         continue;
      }
      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/EXPIRED", 4))
      {
         DoPreExpired = true;
         continue;
      }
      if (strsame (aptr, "/FORMAT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         cptr++;
         if (strsame (cptr, "HEADINGS", 4))
            DoFormatHeadings = true;
         else
         if (strsame (cptr, "NAMES", 4))
            DoFormatNames = true;
         else
         {
            fprintf (stdout, "%%%s-E-IVKEYW, invalid keyword\n \\%s\\\n",
                     Utility, cptr);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         continue;
      }
      if (strsame (aptr, "/LENGTH", 4))
      {
         DoContentLength = true;
         continue;
      }
      if (strsame (aptr, "/LOCATION=", 4))
      {
         DoResponse = true;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliLocationPtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/MAXSYM=", 7))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         SymbolValueMax = atoi(cptr+1);
         continue;
      }
      if (strsame (aptr, "/OUTPUT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliOutputPtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/PREFIX=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         SymbolPrefixPtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/RESPONSE=", 4))
      {
         DoResponse = true;
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (!*cptr) continue;
         CliResponsePtr = cptr+1;
         continue;
      }
      if (strsame (aptr, "/SYMBOLS", 4))
      {
         DoDclSymbols = true;
         continue;
      }
      if (strsame (aptr, "/URLDECODE", 4))
      {
         DoUrlDecode = true;
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }

      if (CliParameterPtr == NULL)
      {
         CliParameterPtr = aptr;
         continue;
      }

      fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
               Utility, aptr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }
}

/*****************************************************************************/
/*
Generate and send to the client an HTTP response header.  Various global
variables feed into the generated header.
*/

SendResponse ()

{
   register char  *cptr;

   int  StatusCode;
   char  *PreExpiredPtr;
   char  ContentLengthHeader [32];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SendResponse()\n");

   if (!Debug)
   {
      /* reopen output stream so that the '\r' and '\n' are not filtered */
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin")) == NULL)
         exit (vaxc$errno);
   }

   if (CliLocationPtr != NULL)
   {
      fprintf (stdout,
"HTTP/1.0 302 %s\r\n\
Server: %s\r\n\
Date: %s\r\n\
Location: %s\r\n\
\r\n",
         HttpStatusCodeText(302),
         CgiServerSoftwarePtr,
         CgiRequestTimeGmtPtr,
         CliLocationPtr);
      return;
   }

   if (FileContentLength >= 0)
      sprintf (ContentLengthHeader, "Content-Length: %d\r\n",
               FileContentLength);
   else
      ContentLengthHeader[0] = '\0';

   if (DoPreExpired)
      PreExpiredPtr = "Expires: Fri, 13 Jan 1978 14:00:00 GMT\r\n";
   else
      PreExpiredPtr = "";

   StatusCode = atoi (CliResponsePtr);
   if (StatusCode < 200 || StatusCode > 599) StatusCode = 0;

   fprintf (stdout,
"HTTP/1.0 %d %s\r\n\
Server: %s\r\n\
Date: %s\r\n\
Content-Type: %s%s\r\n\
%s\
%s\
\r\n",
      StatusCode, HttpStatusCodeText(StatusCode),
      CgiServerSoftwarePtr,
      CgiRequestTimeGmtPtr,
      CliContentTypePtr, CharsetPtr,
      ContentLengthHeader,
      PreExpiredPtr);
}

/*****************************************************************************/
/*
Send a file to the client.  If the file cannot be opened for any reason
exit with error.  If the content-type is text the open it in record-mode,
if not text then in binary mode (this way it should work for files of all
characteristics). If the content-length has been requested count the content
bytes by reading the whole file and rewinding (expensive, but it returns a
valid result even with variable-record-length files).
*/

SendFile ()

{
   int  ReadCount;
   char  Buffer [4096];
   FILE  *InputFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SendFile() |%s|\n", CliParameterPtr);

   if (ResponseContentTypeText)
      InputFile = fopen (CliParameterPtr, "r", "shr=get");
   else
      InputFile = fopen (CliParameterPtr, "r", "ctx=bin", "shr=get");
   if (InputFile == NULL) exit(vaxc$errno);

   if (DoContentLength)
   {
      FileContentLength = 0;
      if (ResponseContentTypeText)
         while (fgets (Buffer, sizeof(Buffer), InputFile) != NULL)
            FileContentLength += strlen(Buffer);
      else
         while (ReadCount = fread (Buffer, 1, sizeof(Buffer), InputFile))
            FileContentLength += ReadCount;
      rewind (InputFile);
   }

   if (DoResponse)
      SendResponse ();
   else
   {
      /* reopen output stream so that the '\r' and '\n' are not filtered */
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin")) == NULL)
         exit (vaxc$errno);
   }


   if (ResponseContentTypeText)
      while (fgets (Buffer, sizeof(Buffer), InputFile) != NULL)
         fputs (Buffer, stdout);
   else
      while (ReadCount = fread (Buffer, 1, sizeof(Buffer), InputFile))
         fwrite (Buffer, ReadCount, 1, stdout);

   fclose (InputFile);
}

/*****************************************************************************/
/*
Return the a pointer to abbreviated meaning of the supplied HTTP status code.
These are typical of those included on the response header status line.
*/

char *HttpStatusCodeText (int StatusCode)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpStatusCodeText() %d\n", StatusCode);

   switch (StatusCode)
   {
      case 000 : return ("Script Error!");
      case 200 : return ("Success");
      case 201 : return ("Created");
      case 202 : return ("Accepted");
      case 203 : return ("No content");
      case 301 : return ("Moved permanently");
      case 302 : return ("Moved temporarily");
      case 304 : return ("Not modified");
      case 400 : return ("Bad request");
      case 401 : return ("Authorization required");
      case 403 : return ("Forbidden");
      case 404 : return ("Not found");
      case 500 : return ("Internal error");
      case 501 : return ("Not implemented");
      case 502 : return ("Bad gateway");
      case 503 : return ("Service unavailable");
      default :  return ("Unknown code!");
   }
}

/*****************************************************************************/
/*
Read a POSTed request body into a single array of char (doesn't matter whether
it's text or binary) pointed to by 'BufferPtr' with a length of 'BufferCount'.
*/

ReadRequestBody ()

{
   int  ReadCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ReadRequestBody()\n");

   /* if the body has alread been read then just return */
   if (BufferPtr != NULL) return;

   GetCgiVar (CgiRequestMethodPtr, "WWW_REQUEST_METHOD");
   if (!strsame (CgiRequestMethodPtr, "POST", -1))
   {
      fprintf (stdout, "%%%s-E-METHOD, is not POST\n \\%s\\\n",
               Utility, CgiRequestMethodPtr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   if ((BufferPtr = calloc (1, BUFFER_CHUNK+1)) == NULL)
      exit (vaxc$errno);
   BufferCount = 0;
   BufferSize = BUFFER_CHUNK;

   while (ReadCount =
          fread (BufferPtr+BufferCount, 1, BufferSize-BufferCount, stdin))
   {
      BufferCount += ReadCount;
      if (BufferCount >= BufferSize)
      {
         if ((BufferPtr =
              realloc (BufferPtr, BufferSize+BUFFER_CHUNK+1)) == NULL)
            exit (vaxc$errno);
         BufferSize += BUFFER_CHUNK;
      }
   }
   if (Debug) fprintf (stdout, "size: %d count: %d\n", BufferSize, BufferCount);

   GetCgiVar (CgiContentTypePtr, "WWW_CONTENT_TYPE");
   if (strsame (CgiContentTypePtr, FORM_URLENCODED, -1) ||
       strsame (CgiContentTypePtr, "text/", 5))
   {
      /* it's text, terminate with a null */
      BufferPtr[BufferCount] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", BufferPtr);
   }
}

/*****************************************************************************/
/*
Write the request body to to the file specified by 'CliOutputPtr'.  Write as
text if form-URL-encoded or text, binary if anything else.
*/

RequestBodyWrite ()

{
   FILE  *OutputFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WriteRequestWrite()\n");

   ReadRequestBody ();

   if (CliOutputPtr == NULL)
      OutputFile = stdout;
   else
   if ((OutputFile = fopen (CliOutputPtr, "w")) == NULL)
      exit (vaxc$errno);

   GetCgiVar (CgiContentTypePtr, "WWW_CONTENT_TYPE");
   if (strsame (CgiContentTypePtr, FORM_URLENCODED, -1))
   {
      UrlDecode (BufferPtr);
      /* write as text file */
      if (fputs (BufferPtr, OutputFile) == EOF)
         exit (vaxc$errno);
   }
   else
   if (strsame (CgiContentTypePtr, "text/", 5))
   {
      /* write as text file */
      if (fputs (BufferPtr, OutputFile) == EOF)
         exit (vaxc$errno);
   }
   else
   {
      /* write as binary */
      if (!fwrite (BufferPtr, 1, BufferCount, OutputFile))
         exit (vaxc$errno);
   }

   fclose (OutputFile);
}

/*****************************************************************************/
/*
Write the form-URL-decoded contents of the request body into an output file. 
Two formats for the output document are supported, names and headings.
*/

UrlDecodeBodyWrite ()

{
   register char  *cptr, *sptr, *zptr;

   int  cnt,
        FieldNameLength;
   char  FieldName [256],
         FieldValue [1024];
   FILE  *OutputFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecodeBodyWrite()\n");

   GetCgiVar (CgiContentTypePtr, "WWW_CONTENT_TYPE");
   if (!strsame (CgiContentTypePtr, FORM_URLENCODED, -1))
   {
      fprintf (stdout, "%%%s-E-CONTENT, is not form-URL-encoded\n \\%s\\\n",
               Utility, CgiContentTypePtr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   ReadRequestBody ();

   if (CliOutputPtr == NULL)
      OutputFile = stdout;
   else
   if ((OutputFile = fopen (CliOutputPtr, "w")) == NULL)
      exit (vaxc$errno);

   cptr = BufferPtr;
   while (*cptr)
   {
      FieldNameLength = 0;
      while (*cptr && *cptr != '=')
      {
         zptr = (sptr = FieldName) + sizeof(FieldName)-1;
         while (*cptr && *cptr != '=' && sptr < zptr)
         {
            if (isprint(*cptr))
               *sptr++ = *cptr++;
            else
               cptr++;
         }
         /* if field name is too large it's just truncated and rest ignored */
         while (*cptr && *cptr != '=') cptr++;
         *sptr = '\0';
         if (Debug) fprintf (stdout, "FieldName: |%s|\n", FieldName);

         UrlDecode (FieldName);
         if (DoFormatHeadings)
            fprintf (OutputFile, "%s\n", FieldName);
         else
            fprintf (OutputFile, "%s: ", FieldName);
         FieldNameLength += sptr - FieldName;
      }
      if (*cptr == '=') cptr++;

      if (DoFormatHeadings)
      {
         /* underline the field name */
         for (cnt = 0; cnt < FieldNameLength; cnt++)
            fputc ('_', OutputFile);
         fputs ("\n", OutputFile);
      }

      while (*cptr && *cptr != '&')
      {
         zptr = (sptr = FieldValue) + sizeof(FieldValue)-1;
         while (*cptr && *cptr != '&' && sptr < zptr)
         {
            if (isprint(*cptr))
               *sptr++ = *cptr++;
            else
               cptr++;
         }
         *sptr = '\0';
         if (Debug) fprintf (stdout, "FieldValue: |%s|\n", FieldValue);

         UrlDecode (FieldValue);
         StripCarRet (FieldValue);
         fputs (FieldValue, OutputFile);
      }
      if (*cptr == '&') cptr++;

      /* blank line between this field and the next */
      fputs ("\n\n", OutputFile);
   }

   fclose (OutputFile);
}

/*****************************************************************************/
/*
Decode the form-URL-encoded POSTed request body with the contents of each field
being placed into DCL symbols named using that field name.  Field contents
greater than the capacity of a single DCL symbol (255 on <=6.n  and 1023 on
>=v7.n) have the contents distributed across multiple symbol names. Each of
these fields comprises the basic name with a dollar symbol and an ascending
integer.  Informational symbols are also created.
*/

UrlDecodeBodyDclSymbols ()

{
   register char  *cptr, *sptr, *zptr;

   boolean  SymbolExceedsMax;
   int  FieldCount,
        SymbolCount,
        SymbolNameLength,
        SymbolValueCount,
        SymbolValueLength,
        SymbolValuePart;
   char  SymbolName [256],
         SymbolValue [1024];
   FILE  *OutputFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecodeBodyDclSymbols()\n");

   GetCgiVar (CgiContentTypePtr, "WWW_CONTENT_TYPE");
   if (!strsame (CgiContentTypePtr, FORM_URLENCODED, -1))
   {
      fprintf (stdout, "%%%s-E-CONTENT, is not form-URL-encoded\n \\%s\\\n",
               Utility, CgiContentTypePtr);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   ReadRequestBody ();

   if (Debug) fprintf (stdout, "|%s|\n", BufferPtr);

   SymbolExceedsMax = false;
   FieldCount = SymbolCount = SymbolNameLength =
      SymbolValueCount = SymbolValueLength = SymbolValuePart = 0;
   SymbolName[0] = SymbolValue[0] = '\0';

   cptr = BufferPtr;
   while (*cptr)
   {
      if (!SymbolNameLength)
      {
         FieldCount++;
         zptr = (sptr = SymbolName) + sizeof(SymbolName);
         memcpy (sptr, SymbolPrefixPtr, SymbolPrefixLength);
         sptr += SymbolPrefixLength;
         if (sptr < zptr) *sptr++ = '_';
         while (*cptr && *cptr != '=' && sptr < zptr)
         {
            if (isprint(*cptr))
               *sptr++ = toupper(*cptr++);
            else
               cptr++;
         }
         if (sptr >= zptr)
         {
            fprintf (stdout, "%%%s-E-SYMBOLNAME, too big\n \\%s\\\n",
                     Utility, SymbolName);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         *sptr = '\0';
         SymbolNameLength = sptr - SymbolName;
         if (Debug) fprintf (stdout, "SymbolName: |%s|\n", SymbolName);

         if (*cptr == '=') cptr++;
      }

      if (!SymbolValueLength)
      {
         zptr = (sptr = SymbolValue) + sizeof(SymbolValue);

         while (*cptr && *cptr != '&' && sptr < zptr &&
                sptr - SymbolValue < SymbolValueMax)
         {
            if (isprint(*cptr))
               *sptr++ = *cptr++;
            else
               cptr++;
         }
         if (sptr >= zptr)
         {
            fprintf (stdout, "%%%s-E-SYMBOLVALUE, too big\n \\%s\\\n",
                     Utility, SymbolName);
            exit (STS$K_ERROR | STS$M_INHIB_MSG);
         }
         *sptr = '\0';
         SymbolValueLength = sptr - SymbolValue;
         if (Debug) fprintf (stdout, "SymbolValue: |%s|\n", SymbolValue);

         if (*cptr == '&')
            cptr++;
         else
         if (*cptr)
            SymbolExceedsMax = true;
         else
            SymbolExceedsMax = false;

         if (SymbolExceedsMax)
         {
            /* symbol value reached maximum size before end of value */
            SymbolValueCount += SymbolValueLength;
            if (SymbolNameLength > 245)
            {
               fprintf (stdout, "%%%s-E-SYMBOLNAME, too big\n \\%s\\\n",
                        Utility, SymbolName);
               exit (STS$K_ERROR | STS$M_INHIB_MSG);
            }
            sprintf (SymbolName+SymbolNameLength, "$%d", ++SymbolValuePart);
            if (Debug) fprintf (stdout, "SymbolName: |%s|\n", SymbolName);
         }
         else
         if (SymbolValueCount)
            SymbolValueCount += SymbolValueLength;
      }

      UrlDecode (SymbolName);
      /* substitute underscores for unacceptable symbol name characters */
      for (sptr = SymbolName; *sptr; sptr++)
         if (!isalnum(*sptr)) *sptr = '_';
      UrlDecode (SymbolValue);
      SymbolCount++;
      if (Debug) fprintf (stdout, "|%s=%s|\n", SymbolName, SymbolValue);

      SetGlobalSymbol (SymbolName, SymbolValue, 0);

      if (SymbolExceedsMax)
      {
         /* symbol value reached maximum size before end of value */
         SymbolValueLength = 0;
         SymbolValue[0] = '\0';
      }
      else
      {
         if (SymbolValueCount)
         {
            strcpy (SymbolName+SymbolNameLength, "$0");
            if (Debug) fprintf (stdout, "SymbolName: |%s|\n", SymbolName);
            SetGlobalSymbol (SymbolName, NULL, SymbolValueCount);
         }

         /* restart both name and value */
         SymbolNameLength = SymbolValueCount =
            SymbolValueLength = SymbolValuePart = 0;
         SymbolName[0] = SymbolValue[0] = '\0';
      }
      SymbolExceedsMax = false;
   }

   sprintf (SymbolName, "%s$FIELDS", SymbolPrefixPtr);
   SetGlobalSymbol (SymbolName, NULL, FieldCount);

   sprintf (SymbolName, "%s$MAXSYM", SymbolPrefixPtr);
   SetGlobalSymbol (SymbolName, NULL, SymbolValueMax);

   sprintf (SymbolName, "%s$SYMBOLS", SymbolPrefixPtr);
   SetGlobalSymbol (SymbolName, NULL, SymbolCount);
}           

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!
*/

int UrlDecode (char *String)

{
   register char  c;
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            c = 0;
            if (*cptr >= '0' && *cptr <= '9')
               c = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               c += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = c;

            continue;

         case '\n' :
         case '\r' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/*****************************************************************************/
/*
Strip carriage-returns, leaving only newlines (for text files).
*/

int StripCarRet (char *String)

{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "StrinCarRet() |%s|\n", String);

   /* don't start actually copying unless we really have to! */
   cptr = String;
   while (*cptr && *cptr != '\r') cptr++;
   sptr = cptr;
   while (*cptr)
   {
      if (*cptr == '\r')
         cptr++;
      else
         *sptr++ = *cptr++;
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/*****************************************************************************/
/*
Get the contents of the specified DCL symbol.  The copy to scratch storage,
adding an extra double-quote for each found in the source symbol.  Reset
original symbol name with the value containing doubled-up quotes.
*/

EscapeSymbolQuotes (char *SymbolName)

{
   register char  *cptr, *eptr, *sptr, *zptr;

   int  status,
        SymbolLength;
   char  SymbolValue [1024],
         DquoteValue [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "EscapeSymbolQuotes() |%s|\n", SymbolName);

   if (VMSnok (status =
       GetGlobalSymbol (SymbolName, SymbolValue, &SymbolLength)))
      exit (status);

   sptr = DquoteValue;
   if (VmsVersion() <= 6)
      zptr = sptr + 255;
   else
      zptr = sptr + 1023;
   eptr = SymbolValue + SymbolLength;
   cptr = SymbolValue;
   while (cptr < eptr && sptr < zptr)
   {
      if (*cptr == '\"' && sptr < zptr) *sptr++ = '\"';
      if (sptr < zptr) *sptr++ = *cptr++;
   }
   /* +1 makes it an error */
   if (sptr >= zptr) exit (SS$_BUFFEROVF+1);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", DquoteValue);

   SetGlobalSymbol (SymbolName, DquoteValue, NULL);
}

/****************************************************************************/
/*
Assign a global symbol.  If the string pointer is null the numeric value is
used.  Symbol lengths are adjusted according to the maximum allowed for the
version of the operating system.
*/ 

SetGlobalSymbol
(
char *Name,
char *String,
int Value
)
{
   static int  GlobalSymbol = LIB$K_CLI_GLOBAL_SYM;
   static char  ValueString [32];
   static $DESCRIPTOR (SymbolValueDsc, "");
   static $DESCRIPTOR (ValueFaoDsc, "!UL");
   static $DESCRIPTOR (ValueStringDsc, ValueString);

   int  status;
   $DESCRIPTOR (SymbolNameDsc, Name);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SetGlobalSymbol() |%s|%s| %d\n",
               Name, String, Value);

   SymbolNameDsc.dsc$w_length = strlen(Name);
   if (String == NULL)
   {
      SymbolValueDsc.dsc$a_pointer = ValueString;
      sys$fao (&ValueFaoDsc, &SymbolValueDsc.dsc$w_length, &ValueStringDsc,
               Value);
      ValueString[SymbolValueDsc.dsc$w_length] = '\0';
   }
   else
   {
      SymbolValueDsc.dsc$a_pointer = String;
      SymbolValueDsc.dsc$w_length = strlen(String);
      if (SymbolValueDsc.dsc$w_length > 255 && VmsVersion() <= 6)
         SymbolValueDsc.dsc$w_length = 255;
      else
      if (SymbolValueDsc.dsc$w_length > 1023)
         SymbolValueDsc.dsc$w_length = 1023;
   }

   if (VMSnok (status =
       lib$set_symbol (&SymbolNameDsc, &SymbolValueDsc, &GlobalSymbol)))
      exit (status);
}

/*****************************************************************************/
/*
Gte a global symbol name's value.  Variable value storage must provide at least
1024 bytes storage.
*/

int GetGlobalSymbol
(
char *VariableName,
char *VariableValue,
int *LengthPtr
)
{
   int  status;
   unsigned short  Length;
   unsigned long  SymbolTable;
   $DESCRIPTOR (VariableNameDsc, VariableName);
   $DESCRIPTOR (VariableValueDsc, VariableValue);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetGlobalSymbol() |%s|\n", VariableName);

   VariableNameDsc.dsc$w_length = strlen(VariableName);
   if (VmsVersion() <= 6)
      VariableValueDsc.dsc$w_length = 255;
   else
      VariableValueDsc.dsc$w_length = 1023;

   status = lib$get_symbol (&VariableNameDsc, &VariableValueDsc, &Length,
                            &SymbolTable);
   if (Debug) fprintf (stdout, "lib$get_symbol() %%X%08.08X\n", status);
   /* if it's not a global symbol then we're noit interested! */
   if (SymbolTable != LIB$K_CLI_GLOBAL_SYM) status = LIB$_NOSUCHSYM;
   if (VMSnok (status))
   {
      VariableValue[0] = '\0';
      return (status);
   }

   if (LengthPtr != NULL) *LengthPtr = Length;
   VariableValue[Length] = '\0';
   if (Debug) fprintf (stdout, "%d |%s|\n", Length, VariableValue);

   return (status);
}

/****************************************************************************/
/*
Return an integer reflecting the major VMS version number.
*/ 

int VmsVersion ()

{
   static char  SyiVersion [16];

   static struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   SyiItems [] =
   {
      { 8, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "VmsVersion()\n");

   if (SyiVersion[0]) return (SyiVersion[1]-30);

   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      exit (status);
   SyiVersion[8] = '\0';
   if (Debug) fprintf (stdout, "SyiVersion |%s|\n", SyiVersion);
   return (SyiVersion[1]-30);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/

